'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

const DefaultFetchAPI = require('@whatwg-node/fetch');
const server = require('@whatwg-node/server');

const HTTP_METHODS = [
    'get',
    'post',
    'put',
    'delete',
    'head',
    'options',
    'patch',
    'trace',
];
const Response = DefaultFetchAPI.Response;
function createRouterBase({ fetchAPI: givenFetchAPI, base: basePath = '/', plugins = [], } = {}) {
    const fetchAPI = {
        ...DefaultFetchAPI,
        ...givenFetchAPI,
    };
    const onRouterInit = [];
    const onRouteHooks = [];
    for (const plugin of plugins) {
        if (plugin.onRouterInit) {
            onRouterInit.push(plugin.onRouterInit);
        }
        if (plugin.onRoute) {
            onRouteHooks.push(plugin.onRoute);
        }
    }
    const routesByMethod = new Map();
    function addHandlersToMethod({ operationId, description, method, path, schemas, handlers, }) {
        for (const onRouteHook of onRouteHooks) {
            onRouteHook({
                operationId,
                description,
                method,
                path,
                schemas,
                handlers,
            });
        }
        let methodPatternMaps = routesByMethod.get(method);
        if (!methodPatternMaps) {
            methodPatternMaps = new Map();
            routesByMethod.set(method, methodPatternMaps);
        }
        let fullPath = '';
        if (basePath === '/') {
            fullPath = path;
        }
        else if (path === '/') {
            fullPath = basePath;
        }
        else {
            fullPath = `${basePath}${path}`;
        }
        const pattern = new fetchAPI.URLPattern({ pathname: fullPath });
        methodPatternMaps.set(pattern, handlers);
    }
    async function handleRequest(request, context) {
        const method = request.method.toLowerCase();
        let _parsedUrl;
        function getParsedUrl() {
            if (!_parsedUrl) {
                _parsedUrl = new fetchAPI.URL(request.url);
            }
            return _parsedUrl;
        }
        const methodPatternMaps = routesByMethod.get(method);
        if (methodPatternMaps) {
            const queryProxy = new Proxy({}, {
                get(_, prop) {
                    const parsedUrl = getParsedUrl();
                    const allQueries = parsedUrl.searchParams.getAll(prop.toString());
                    return allQueries.length === 1 ? allQueries[0] : allQueries;
                },
                has(_, prop) {
                    const parsedUrl = getParsedUrl();
                    return parsedUrl.searchParams.has(prop.toString());
                },
            });
            for (const [pattern, handlers] of methodPatternMaps) {
                const match = pattern.exec(request.url);
                if (match) {
                    const routerRequest = new Proxy(request, {
                        get(target, prop) {
                            if (prop === 'parsedUrl') {
                                return getParsedUrl();
                            }
                            if (prop === 'params') {
                                return new Proxy(match.pathname.groups, {
                                    get(_, prop) {
                                        const value = match.pathname.groups[prop.toString()];
                                        if (value != null) {
                                            return decodeURIComponent(value);
                                        }
                                        return value;
                                    },
                                });
                            }
                            if (prop === 'query') {
                                return queryProxy;
                            }
                            const targetProp = target[prop];
                            if (typeof targetProp === 'function') {
                                return targetProp.bind(target);
                            }
                            return targetProp;
                        },
                        has(target, prop) {
                            return (prop in target || prop === 'parsedUrl' || prop === 'params' || prop === 'query');
                        },
                    });
                    for (const handler of handlers) {
                        const result = await handler(routerRequest, context);
                        if (result) {
                            return result;
                        }
                    }
                }
            }
        }
    }
    const router = new Proxy({}, {
        get(_, prop) {
            if (prop === 'handle') {
                return handleRequest;
            }
            if (prop === 'addRoute') {
                return function (opts) {
                    const { operationId, description, method, path, schemas, handler } = opts;
                    addHandlersToMethod({
                        operationId,
                        description,
                        method,
                        path,
                        schemas,
                        handlers: [handler],
                    });
                    return this;
                };
            }
            const method = prop.toString().toLowerCase();
            return function routeMethodKeyFn(path, ...handlers) {
                if (method === 'all') {
                    for (const httpMethod of HTTP_METHODS) {
                        addHandlersToMethod({
                            method: httpMethod.toLowerCase(),
                            path,
                            handlers,
                        });
                    }
                }
                else {
                    addHandlersToMethod({
                        method: method.toLowerCase(),
                        path,
                        handlers,
                    });
                }
                return this;
            };
        },
    });
    for (const onRouterInitHook of onRouterInit) {
        onRouterInitHook(router);
    }
    return router;
}
function createRouter(options) {
    const routerBaseObject = createRouterBase(options);
    return server.createServerAdapter(routerBaseObject, options);
}

const swaggerUiHtml = "<!DOCTYPE html><html lang=en><head><meta charset=utf-8><meta name=viewport content=\"width=device-width,initial-scale=1\"><meta name=description content=SwaggerUI><title>SwaggerUI</title><link rel=stylesheet href=https://unpkg.com/swagger-ui-dist/swagger-ui.css></head><body><div id=swagger-ui></div><script src=https://unpkg.com/swagger-ui-dist/swagger-ui-bundle.js crossorigin></script><script>window.onload=()=>{window.ui=SwaggerUIBundle({url:__OAS_PATH__,dom_id:\"#swagger-ui\",displayOperationId:!0})}</script></body></html>";

function useOpenAPI({ oasPath = '/openapi.json', swaggerUIPath = '/docs', baseOas: oas = {
    openapi: '3.0.1',
    info: {
        title: 'My API',
        version: '1.0.0',
    },
    components: {},
}, } = {}) {
    const paths = (oas.paths || (oas.paths = {}));
    return {
        onRouterInit(router) {
            router.get(oasPath, () => Response.json(oas));
            router.get(swaggerUIPath, () => new Response(swaggerUiHtml.replace('__OAS_PATH__', JSON.stringify(oasPath)), {
                headers: {
                    'Content-Type': 'text/html',
                },
                status: 200,
            }));
        },
        onRoute({ method, path, operationId, description, schemas }) {
            var _a, _b, _c, _d, _e, _f, _g;
            if (schemas) {
                const pathObj = (paths[path] = paths[path] || {});
                pathObj[method] = (pathObj[method] || {});
                const operation = pathObj[method];
                operation.operationId = operationId;
                operation.description = description;
                if (schemas.Responses) {
                    for (const statusCode in schemas.Responses) {
                        const response = schemas.Responses[statusCode];
                        operation.responses = operation.responses || {};
                        operation.responses[statusCode] = {
                            description: '',
                            content: {
                                'application/json': {
                                    schema: response,
                                },
                            },
                        };
                    }
                }
                if (((_a = schemas.Request) === null || _a === void 0 ? void 0 : _a.Headers) &&
                    typeof schemas.Request.Headers === 'object' &&
                    'properties' in schemas.Request.Headers) {
                    for (const headerName in schemas.Request.Headers.properties) {
                        const headersSchema = schemas.Request.Headers.properties[headerName];
                        operation.parameters = operation.parameters || [];
                        operation.parameters.push({
                            name: headerName,
                            in: 'header',
                            required: (_b = schemas.Request.Headers.required) === null || _b === void 0 ? void 0 : _b.includes(headerName),
                            schema: headersSchema,
                        });
                    }
                }
                if (((_c = schemas.Request) === null || _c === void 0 ? void 0 : _c.PathParams) &&
                    typeof schemas.Request.PathParams === 'object' &&
                    'properties' in schemas.Request.PathParams) {
                    for (const paramName in schemas.Request.PathParams.properties) {
                        const paramSchema = schemas.Request.PathParams.properties[paramName];
                        operation.parameters = operation.parameters || [];
                        operation.parameters.push({
                            name: paramName,
                            in: 'path',
                            required: (_d = schemas.Request.PathParams.required) === null || _d === void 0 ? void 0 : _d.includes(paramName),
                            schema: paramSchema,
                        });
                    }
                }
                if (((_e = schemas.Request) === null || _e === void 0 ? void 0 : _e.QueryParams) &&
                    typeof schemas.Request.QueryParams === 'object' &&
                    'properties' in schemas.Request.QueryParams) {
                    for (const paramName in schemas.Request.QueryParams.properties) {
                        const paramSchema = schemas.Request.QueryParams.properties[paramName];
                        operation.parameters = operation.parameters || [];
                        operation.parameters.push({
                            name: paramName,
                            in: 'query',
                            required: (_f = schemas.Request.QueryParams.required) === null || _f === void 0 ? void 0 : _f.includes(paramName),
                            schema: paramSchema,
                        });
                    }
                }
                if ((_g = schemas.Request) === null || _g === void 0 ? void 0 : _g.JSONBody) {
                    operation.requestBody = {
                        content: {
                            'application/json': {
                                schema: schemas.Request.JSONBody,
                            },
                        },
                    };
                }
            }
        },
    };
}

Object.defineProperty(exports, 'URLPattern', {
    enumerable: true,
    get: function () {
        return DefaultFetchAPI.URLPattern;
    }
});
Object.defineProperty(exports, 'useCORS', {
    enumerable: true,
    get: function () {
        return server.useCORS;
    }
});
Object.defineProperty(exports, 'useErrorHandling', {
    enumerable: true,
    get: function () {
        return server.useErrorHandling;
    }
});
exports.Response = Response;
exports.createRouter = createRouter;
exports.createRouterBase = createRouterBase;
exports.useOpenAPI = useOpenAPI;
